"""DO NOT MODIFY. Auto-generated by script/fingerprint_frontend."""

FINGERPRINTS = {
    "core.js": "5ed5e063d66eb252b5b288738c9c2d16",
    "frontend.html": "78be2dfedc4e95326cbcd9401fb17b4d",
    "mdi.html": "46a76f877ac9848899b8ed382427c16f",
    "micromarkdown-js.html": "93b5ec4016f0bba585521cf4d18dec1a",
    "panels/ha-panel-dev-event.html": "550bf85345c454274a40d15b2795a002",
    "panels/ha-panel-dev-info.html": "ec613406ce7e20d93754233d55625c8a",
    "panels/ha-panel-dev-service.html": "4a051878b92b002b8b018774ba207769",
    "panels/ha-panel-dev-state.html": "65e5f791cc467561719bf591f1386054",
    "panels/ha-panel-dev-template.html": "7d744ab7f7c08b6d6ad42069989de400",
    "panels/ha-panel-history.html": "efe1bcdd7733b09e55f4f965d171c295",
    "panels/ha-panel-iframe.html": "d920f0aa3c903680f2f8795e2255daab",
    "panels/ha-panel-logbook.html": "66108d82763359a218c9695f0553de40",
    "panels/ha-panel-map.html": "49ab2d6f180f8bdea7cffaa66b8a5d3e"
}
